from __future__ import annotations

import importlib
import traceback
import typing as t
from contextlib import ExitStack
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING

from ..._internal.utils import LazyLoader
from ..._internal.utils import cached_contextmanager
from ..._internal.utils import reserve_free_port
from ...exceptions import BentoMLException
from ...grpc.utils import LATEST_PROTOCOL_VERSION
from ...grpc.utils import import_generated_stubs
from ...grpc.utils import import_grpc

if TYPE_CHECKING:
    import grpc
    import numpy as np
    from google.protobuf.message import Message
    from grpc import aio
    from grpc.aio._channel import Channel
    from numpy.typing import NDArray

    from ..._internal.service import Service
    from ...grpc.v1 import service_pb2 as pb
else:
    grpc, aio = import_grpc()  # pylint: disable=E1111
    np = LazyLoader("np", globals(), "numpy")

__all__ = [
    "async_client_call",
    "randomize_pb_ndarray",
    "make_pb_ndarray",
    "create_channel",
    "make_standalone_server",
    "create_test_bento_servicer",
]


def create_test_bento_servicer(
    service: Service,
    protocol_version: str = LATEST_PROTOCOL_VERSION,
) -> t.Callable[[Service], t.Any]:
    try:
        module = importlib.import_module(
            f".{protocol_version}", package="bentoml._internal.server.grpc.servicer"
        )
        return getattr(module, "create_bento_servicer")(service)
    except (ImportError, ModuleNotFoundError):
        raise BentoMLException(
            f"Failed to load servicer implementation for version {protocol_version}"
        ) from None


def randomize_pb_ndarray(
    shape: tuple[int, ...], protocol_version: str = LATEST_PROTOCOL_VERSION
) -> pb.NDArray:
    pb, _ = import_generated_stubs(protocol_version)
    arr: NDArray[np.float32] = t.cast("NDArray[np.float32]", np.random.rand(*shape))
    return pb.NDArray(
        shape=list(shape), dtype=pb.NDArray.DTYPE_FLOAT, float_values=arr.ravel()
    )


def make_pb_ndarray(
    arr: NDArray[t.Any], protocol_version: str = LATEST_PROTOCOL_VERSION
) -> pb.NDArray:
    from bentoml._internal.io_descriptors.numpy import npdtype_to_dtypepb_map
    from bentoml._internal.io_descriptors.numpy import npdtype_to_fieldpb_map

    pb, _ = import_generated_stubs(protocol_version)

    try:
        fieldpb = npdtype_to_fieldpb_map()[arr.dtype]
        dtypepb = npdtype_to_dtypepb_map()[arr.dtype]
        return pb.NDArray(
            **{
                fieldpb: arr.ravel().tolist(),
                "dtype": dtypepb,
                "shape": tuple(arr.shape),
            },
        )
    except KeyError:
        raise BentoMLException(
            f"Unsupported dtype '{arr.dtype}' for response message.",
        ) from None


async def async_client_call(
    method: str,
    channel: Channel,
    data: dict[str, Message | pb.Part | bytes | str | dict[str, t.Any]],
    sanity: bool = True,
    timeout: int | None = 90,
    assert_data: pb.Response | t.Callable[[pb.Response], bool] | None = None,
    assert_code: grpc.StatusCode | None = None,
    assert_details: str | None = None,
    assert_trailing_metadata: aio.Metadata | None = None,
    protocol_version: str = LATEST_PROTOCOL_VERSION,
) -> pb.Response | None:
    """
    Invoke a given API method via a client.

    Args:
        method: The method name to call.
        channel: The given aio.Channel to use for invoking the RPC. Channels shouldn't include
                 any client interceptors. as we will handle interceptors' lifecycle separately.
        data: The data to send to the server.
        assert_data: The data to assert against the response.
        timeout: The timeout for the RPC.
        sanity: Whether to perform sanity check on the response.
        assert_code: The code to assert against the response.
        assert_details: The details to assert against the response.

    Returns:
        The response from the server.
    """
    pb, _ = import_generated_stubs(protocol_version)

    res: pb.Response | None = None
    try:
        Call = channel.unary_unary(
            f"/bentoml.grpc.{protocol_version}.BentoService/Call",
            request_serializer=pb.Request.SerializeToString,
            response_deserializer=pb.Response.FromString,
        )
        output: aio.UnaryUnaryCall[pb.Request, pb.Response] = Call(
            pb.Request(api_name=method, **data), timeout=timeout
        )
        res = await t.cast(t.Awaitable[pb.Response], output)
        return_code = await output.code()
        details = await output.details()
        trailing_metadata = await output.trailing_metadata()
        if sanity:
            assert isinstance(res, pb.Response)
        if assert_data:
            if callable(assert_data):
                assert assert_data(res), f"Failed while checking data: {output}"
            else:
                assert res == assert_data, f"Failed while checking data: {output}"
    except aio.AioRpcError as call:
        return_code = call.code()
        details = call.details()
        trailing_metadata = call.trailing_metadata()
    if assert_code is not None:
        assert (
            return_code == assert_code
        ), f"Method '{method}' returns {return_code} while expecting {assert_code}."
    if assert_details is not None:
        assert (
            assert_details == details
        ), f"Details '{assert_details}' is not in '{details}'."
    if assert_trailing_metadata is not None:
        assert (
            trailing_metadata == assert_trailing_metadata
        ), f"Trailing metadata '{trailing_metadata}' while expecting '{assert_trailing_metadata}'."
    return res


@asynccontextmanager
async def create_channel(
    host_url: str,
    interceptors: t.Sequence[aio.ClientInterceptor] | None = None,
) -> t.AsyncGenerator[Channel, None]:
    """
    Create an async channel with given host_url and client interceptors.

    Args:
        host_url: The host url to connect to.
        interceptors: The client interceptors to use. This is optional, by default set to None.

    Returns:
        A insecure channel.
    """
    channel: Channel | None = None
    try:
        async with aio.insecure_channel(host_url, interceptors=interceptors) as channel:
            # create a blocking call to wait til channel is ready.
            await channel.channel_ready()
            yield channel
    except aio.AioRpcError as e:
        traceback.print_exc()
        raise e from None
    finally:
        if channel:
            await channel.close()


@cached_contextmanager("{interceptors}")
def make_standalone_server(
    interceptors: t.Sequence[aio.ServerInterceptor] | None = None,
    host: str = "0.0.0.0",
) -> t.Generator[tuple[aio.Server, str], None, None]:
    """
    Create a standalone aio.Server for testing.

    Args:
        interceptors: The interceptors to use for the server, default to None.

    Returns:
        The server and the host_url.

    Example for async test cases:

    .. code-block:: python

        async def test_some_async():
            with make_standalone_server() as (server, host_url):
                try:
                    await server.start()
                    channel = grpc.aio.insecure_channel(host_url)
                    ...  # test code here
                finally:
                    await server.stop(None)

    Example for sync test cases:

    .. code-block:: python

        def test_cases():
            import asyncio

            loop = asyncio.new_event_loop()
            with make_standalone_server() as (server, host_url):
                try:
                    loop.run_until_complete(server.start())
                    channel = grpc.insecure_channel(host_url)
                    ...  # test code here
                finally:
                    loop.call_soon_threadsafe(lambda: asyncio.ensure_future(server.stop(None)))
                    loop.close()
                assert loop.is_closed()
    """
    stack = ExitStack()
    port = stack.enter_context(reserve_free_port(enable_so_reuseport=True))
    server = aio.server(
        interceptors=interceptors,
        options=(("grpc.so_reuseport", 1),),
    )
    server.add_insecure_port(f"{host}:{port}")
    print("Using port %d..." % port)
    try:
        yield server, "localhost:%d" % port
    finally:
        stack.close()



from __future__ import annotations

import logging
import typing as t
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING

import cloudpickle

import bentoml
from bentoml import Tag

from ...exceptions import NotFound
from ..models import Model
from ..models.model import ModelContext
from ..models.model import PartialKwargsModelOptions as ModelOptions
from ..types import LazyType
from ..utils.pkg import get_pkg_version
from .common.pytorch import PyTorchTensorContainer
from .common.pytorch import torch

__all__ = ["load_model", "save_model", "get_runnable", "get", "PyTorchTensorContainer"]


MODULE_NAME = "bentoml.pytorch"
MODEL_FILENAME = "saved_model.pt"
API_VERSION = "v1"

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from ..models.model import ModelSignaturesType


def get(tag_like: str | Tag) -> Model:
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, not loading with {MODULE_NAME}."
        )
    return model


def load_model(
    bentoml_model: str | Tag | Model,
    device_id: t.Optional[str] = "cpu",
) -> torch.nn.Module:
    """
    Load a model from a BentoML Model with given name.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        device_id (:code:`str`, `optional`, default to :code:`cpu`):
            Optional devices to put the given model on. Refer to `device attributes <https://pytorch.org/docs/stable/tensor_attributes.html#torch.torch.device>`_.

    Returns:
        :obj:`torch.nn.Module`: an instance of :code:`torch.nn.Module` from BentoML modelstore.

    Examples:

    .. code-block:: python

        import bentoml
        model = bentoml.pytorch.load_model('lit_classifier:latest', device_id="cuda:0")
    """
    if isinstance(bentoml_model, (str, Tag)):
        bentoml_model = get(bentoml_model)

    if bentoml_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bentoml_model.tag} was saved with module {bentoml_model.info.module}, not loading with {MODULE_NAME}."
        )

    weight_file = bentoml_model.path_of(MODEL_FILENAME)
    with Path(weight_file).open("rb") as file:
        model: "torch.nn.Module" = torch.load(file, map_location=device_id)
    return model


def save_model(
    name: Tag | str,
    model: "torch.nn.Module",
    *,
    signatures: ModelSignaturesType | None = None,
    labels: t.Dict[str, str] | None = None,
    custom_objects: t.Dict[str, t.Any] | None = None,
    external_modules: t.List[ModuleType] | None = None,
    metadata: t.Dict[str, t.Any] | None = None,
) -> bentoml.Model:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (:code:`str`):
            Name for given model instance. This should pass Python identifier check.
        model (:code:`torch.nn.Module`):
            Instance of model to be saved
        signatures (:code:`ModelSignaturesType`, `optional`, default to :code:`None`):
            A dictionary of method names and their corresponding signatures.
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
            e.g. a tokenizer instance, preprocessor function, model configuration json
        external_modules (:code:`List[ModuleType]`, `optional`, default to :code:`None`):
            user-defined additional python modules to be saved alongside the model or custom objects,
            e.g. a tokenizer module, preprocessor module, model configuration module
        metadata (:code:`Dict[str, Any]`, `optional`,  default to :code:`None`):
            Custom metadata for given model.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the user-defined model's name, and a generated `version` by BentoML.

    Examples:

    .. code-block:: python

        import torch
        import bentoml

        class NGramLanguageModeler(nn.Module):

            def __init__(self, vocab_size, embedding_dim, context_size):
                super(NGramLanguageModeler, self).__init__()
                self.embeddings = nn.Embedding(vocab_size, embedding_dim)
                self.linear1 = nn.Linear(context_size * embedding_dim, 128)
                self.linear2 = nn.Linear(128, vocab_size)

            def forward(self, inputs):
                embeds = self.embeddings(inputs).view((1, -1))
                out = F.relu(self.linear1(embeds))
                out = self.linear2(out)
                log_probs = F.log_softmax(out, dim=1)
                return log_probs

        tag = bentoml.pytorch.save("ngrams", NGramLanguageModeler(len(vocab), EMBEDDING_DIM, CONTEXT_SIZE))
        # example tag: ngrams:20201012_DE43A2

    Integration with Torch Hub and BentoML:

    .. code-block:: python

        import torch
        import bentoml

        resnet50 = torch.hub.load("pytorch/vision", "resnet50", pretrained=True)
        ...
        # trained a custom resnet50

        tag = bentoml.pytorch.save("resnet50", resnet50)
    """
    if not LazyType("torch.nn.Module").isinstance(model):
        raise TypeError(f"Given model ({model}) is not a torch.nn.Module.")

    context: ModelContext = ModelContext(
        framework_name="torch",
        framework_versions={"torch": get_pkg_version("torch")},
    )

    if signatures is None:
        signatures = {"__call__": {"batchable": False}}
        logger.info(
            'Using the default model signature for PyTorch (%s) for model "%s".',
            signatures,
            name,
        )

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        api_version=API_VERSION,
        labels=labels,
        signatures=signatures,
        custom_objects=custom_objects,
        external_modules=external_modules,
        options=ModelOptions(),
        context=context,
        metadata=metadata,
    ) as bento_model:
        weight_file = bento_model.path_of(MODEL_FILENAME)
        with open(weight_file, "wb") as file:
            torch.save(model, file, pickle_module=cloudpickle)  # type: ignore

        return bento_model


def get_runnable(bento_model: Model):
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """
    from .common.pytorch import PytorchModelRunnable
    from .common.pytorch import make_pytorch_runnable_method
    from .common.pytorch import partial_class

    partial_kwargs: t.Dict[str, t.Any] = bento_model.info.options.partial_kwargs  # type: ignore

    runnable_class: type[PytorchModelRunnable] = partial_class(
        PytorchModelRunnable,
        bento_model=bento_model,
        loader=load_model,
    )
    for method_name, options in bento_model.info.signatures.items():
        method_partial_kwargs = partial_kwargs.get(method_name)
        runnable_class.add_method(
            make_pytorch_runnable_method(method_name, method_partial_kwargs),
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )
    return runnable_class



from __future__ import annotations

import functools
import logging
import typing as t
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING

from opentelemetry import trace
from opentelemetry.context import attach
from opentelemetry.context import detach
from opentelemetry.propagate import extract
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace.status import Status
from opentelemetry.trace.status import StatusCode
from simple_di import Provide
from simple_di import inject

from bentoml._internal.configuration.containers import BentoMLContainer
from bentoml._internal.utils.pkg import get_pkg_version
from bentoml.grpc.utils import GRPC_CONTENT_TYPE
from bentoml.grpc.utils import import_grpc
from bentoml.grpc.utils import parse_method_name
from bentoml.grpc.utils import wrap_rpc_handler

if TYPE_CHECKING:
    import grpc
    from grpc import aio
    from grpc.aio._typing import MetadataKey
    from grpc.aio._typing import MetadataType
    from grpc.aio._typing import MetadataValue
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.trace import Span

    from bentoml.grpc.types import AsyncHandlerMethod
    from bentoml.grpc.types import BentoServicerContext
    from bentoml.grpc.types import HandlerCallDetails
    from bentoml.grpc.types import Request
    from bentoml.grpc.types import Response
    from bentoml.grpc.types import RpcMethodHandler
else:
    grpc, aio = import_grpc()

logger = logging.getLogger(__name__)


class _OpenTelemetryServicerContext(aio.ServicerContext["Request", "Response"]):
    def __init__(self, servicer_context: BentoServicerContext, active_span: Span):
        self._servicer_context = servicer_context
        self._active_span = active_span
        self._code = grpc.StatusCode.OK
        self._details = ""
        super().__init__()

    def __getattr__(self, attr: str) -> t.Any:
        return getattr(self._servicer_context, attr)

    async def read(self) -> Request:
        return await self._servicer_context.read()

    async def write(self, message: Response) -> None:
        return await self._servicer_context.write(message)

    def trailing_metadata(self) -> aio.Metadata:
        return self._servicer_context.trailing_metadata()  # type: ignore (unfinished type)

    def auth_context(self) -> t.Mapping[str, t.Iterable[bytes]]:
        return self._servicer_context.auth_context()

    def peer_identity_key(self) -> str | None:
        return self._servicer_context.peer_identity_key()

    def peer_identities(self) -> t.Iterable[bytes] | None:
        return self._servicer_context.peer_identities()

    def peer(self) -> str:
        return self._servicer_context.peer()

    def disable_next_message_compression(self) -> None:
        self._servicer_context.disable_next_message_compression()

    def set_compression(self, compression: grpc.Compression) -> None:
        return self._servicer_context.set_compression(compression)

    def invocation_metadata(self) -> aio.Metadata | None:
        return self._servicer_context.invocation_metadata()

    def set_trailing_metadata(self, trailing_metadata: MetadataType) -> None:
        self._servicer_context.set_trailing_metadata(trailing_metadata)

    async def send_initial_metadata(self, initial_metadata: MetadataType) -> None:
        return await self._servicer_context.send_initial_metadata(initial_metadata)

    async def abort(
        self,
        code: grpc.StatusCode,
        details: str = "",
        trailing_metadata: MetadataType = tuple(),
    ) -> None:
        self._code = code
        self._details = details
        self._active_span.set_attribute(
            SpanAttributes.RPC_GRPC_STATUS_CODE, code.value[0]
        )
        self._active_span.set_status(
            Status(status_code=StatusCode.ERROR, description=f"{code}:{details}")
        )
        return await self._servicer_context.abort(
            code, details=details, trailing_metadata=trailing_metadata
        )

    def set_code(self, code: grpc.StatusCode) -> None:
        self._code = code
        details = self._details or code.value[1]
        self._active_span.set_attribute(
            SpanAttributes.RPC_GRPC_STATUS_CODE, code.value[0]
        )
        if code != grpc.StatusCode.OK:
            self._active_span.set_status(
                Status(status_code=StatusCode.ERROR, description=f"{code}:{details}")
            )
        return self._servicer_context.set_code(code)

    def code(self) -> grpc.StatusCode:
        return self._code

    def set_details(self, details: str) -> None:
        self._details = details
        if self._code != grpc.StatusCode.OK:
            self._active_span.set_status(
                Status(
                    status_code=StatusCode.ERROR, description=f"{self._code}:{details}"
                )
            )
        return self._servicer_context.set_details(details)

    def details(self) -> str:
        return self._details


# Since opentelemetry doesn't provide an async implementation for the server interceptor,
# we will need to create an async implementation ourselves.
# By doing this we will have more control over how to handle span and context propagation.
#
# Until there is a solution upstream, this implementation is sufficient for our needs.
class AsyncOpenTelemetryServerInterceptor(aio.ServerInterceptor):
    @inject
    def __init__(
        self,
        *,
        tracer_provider: TracerProvider = Provide[BentoMLContainer.tracer_provider],
        schema_url: str | None = None,
    ):
        self._tracer = tracer_provider.get_tracer(
            "opentelemetry.instrumentation.grpc",
            get_pkg_version("opentelemetry-instrumentation-grpc"),
            schema_url=schema_url,
        )

    @asynccontextmanager
    async def set_remote_context(
        self, servicer_context: BentoServicerContext
    ) -> t.AsyncGenerator[None, None]:
        metadata = servicer_context.invocation_metadata()
        if metadata:
            md: dict[MetadataKey, MetadataValue] = {m.key: m.value for m in metadata}
            ctx = extract(md)
            token = attach(ctx)
            try:
                yield
            finally:
                detach(token)
        else:
            yield

    def start_span(
        self,
        method_name: str,
        context: BentoServicerContext,
        set_status_on_exception: bool = False,
    ) -> t.ContextManager[Span]:
        attributes: dict[str, str | bytes] = {
            SpanAttributes.RPC_SYSTEM: "grpc",
            SpanAttributes.RPC_GRPC_STATUS_CODE: grpc.StatusCode.OK.value[0],
        }

        # method_name shouldn't be none, otherwise
        # it will never reach this point.
        method_rpc, _ = parse_method_name(method_name)
        attributes.update(
            {
                SpanAttributes.RPC_METHOD: method_rpc.method,
                SpanAttributes.RPC_SERVICE: method_rpc.fully_qualified_service,
            }
        )

        # add some attributes from the metadata
        metadata = context.invocation_metadata()
        if metadata:
            dct: dict[str, str | bytes] = dict(metadata)
            if "user-agent" in dct:
                attributes["rpc.user_agent"] = dct["user-agent"]

        # get trailing metadata
        trailing_metadata: MetadataType | None = context.trailing_metadata()
        if trailing_metadata:
            trailing = dict(trailing_metadata)
            attributes["rpc.content_type"] = trailing.get(
                "content-type", GRPC_CONTENT_TYPE
            )

        # Split up the peer to keep with how other telemetry sources
        # do it. This looks like:
        # * ipv6:[::1]:57284
        # * ipv4:127.0.0.1:57284
        # * ipv4:10.2.1.1:57284,127.0.0.1:57284
        #
        # the process ip and port would be [::1] 57284
        try:
            ipv4_addr = context.peer().split(",")[0]
            ip, port = ipv4_addr.split(":", 1)[1].rsplit(":", 1)
            attributes.update(
                {
                    SpanAttributes.NET_PEER_IP: ip,
                    SpanAttributes.NET_PEER_PORT: port,
                }
            )
            # other telemetry sources add this, so we will too
            if ip in ("[::1]", "127.0.0.1"):
                attributes[SpanAttributes.NET_PEER_NAME] = "localhost"
        except IndexError:
            logger.warning("Failed to parse peer address '%s'", context.peer())

        return self._tracer.start_as_current_span(
            name=method_name,
            kind=trace.SpanKind.SERVER,
            attributes=attributes,
            set_status_on_exception=set_status_on_exception,
        )

    async def intercept_service(
        self,
        continuation: t.Callable[[HandlerCallDetails], t.Awaitable[RpcMethodHandler]],
        handler_call_details: HandlerCallDetails,
    ) -> RpcMethodHandler:
        handler = await continuation(handler_call_details)
        method_name = handler_call_details.method

        # Currently not support streaming RPCs.
        if handler and (handler.response_streaming or handler.request_streaming):
            return handler

        def wrapper(behaviour: AsyncHandlerMethod[Response]):
            @functools.wraps(behaviour)
            async def new_behaviour(
                request: Request, context: BentoServicerContext
            ) -> Response | t.Awaitable[Response]:
                async with self.set_remote_context(context):
                    with self.start_span(method_name, context) as span:
                        # wrap context
                        wrapped_context = _OpenTelemetryServicerContext(context, span)

                        # And now we run the actual RPC.
                        try:
                            return await behaviour(request, wrapped_context)
                        except Exception as e:
                            # We are interested in uncaught exception, otherwise
                            # it will be handled by gRPC.
                            if type(e) != Exception:
                                span.record_exception(e)
                            raise e

            return new_behaviour

        return wrap_rpc_handler(wrapper, handler)



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

import anyio
from starlette.responses import PlainTextResponse

if TYPE_CHECKING:
    from ... import external_typing as ext


class TimeoutMiddleware:
    def __init__(self, app: ext.ASGIApp, timeout: int) -> None:
        self.app = app
        self.timeout = timeout

    async def __call__(
        self, scope: ext.ASGIScope, receive: ext.ASGIReceive, send: ext.ASGISend
    ) -> None:
        if scope["type"] not in ("http", "websocket"):
            return await self.app(scope, receive, send)
        async with anyio.create_task_group():
            try:
                with anyio.fail_after(self.timeout):
                    await self.app(scope, receive, send)
            except TimeoutError:
                resp = PlainTextResponse(
                    f"Not able to process the request in {self.timeout} seconds",
                    status_code=504,
                )
                await resp(scope, receive, send)


class MaxConcurrencyMiddleware:
    def __init__(self, app: ext.ASGIApp, max_concurrency: int) -> None:
        self.app = app
        self._semaphore = asyncio.Semaphore(max_concurrency)

    async def __call__(
        self, scope: ext.ASGIScope, receive: ext.ASGIReceive, send: ext.ASGISend
    ) -> None:
        if scope["type"] not in ("http", "websocket"):
            return await self.app(scope, receive, send)

        if self._semaphore.locked():
            resp = PlainTextResponse("Too many requests", status_code=429)
            await resp(scope, receive, send)
            return

        async with self._semaphore:
            await self.app(scope, receive, send)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import builtins
import collections.abc
import concurrent.futures
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.service
import google.protobuf.struct_pb2
import google.protobuf.wrappers_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Request(google.protobuf.message.Message):
    """Request message for incoming Call."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    API_NAME_FIELD_NUMBER: builtins.int
    NDARRAY_FIELD_NUMBER: builtins.int
    DATAFRAME_FIELD_NUMBER: builtins.int
    SERIES_FIELD_NUMBER: builtins.int
    FILE_FIELD_NUMBER: builtins.int
    TEXT_FIELD_NUMBER: builtins.int
    JSON_FIELD_NUMBER: builtins.int
    MULTIPART_FIELD_NUMBER: builtins.int
    SERIALIZED_BYTES_FIELD_NUMBER: builtins.int
    api_name: builtins.str
    """api_name defines the API entrypoint to call.
    api_name is the name of the function defined in bentoml.Service.
    Example:

        @svc.api(input=NumpyNdarray(), output=File())
        def predict(input: NDArray[float]) -> bytes:
            ...

        api_name is "predict" in this case.
    """
    @property
    def ndarray(self) -> global___NDArray:
        """NDArray represents a n-dimensional array of arbitrary type."""
    @property
    def dataframe(self) -> global___DataFrame:
        """DataFrame represents any tabular data type. We are using
        DataFrame as a trivial representation for tabular type.
        """
    @property
    def series(self) -> global___Series:
        """Series portrays a series of values. This can be used for
        representing Series types in tabular data.
        """
    @property
    def file(self) -> global___File:
        """File represents for any arbitrary file type. This can be
        plaintext, image, video, audio, etc.
        """
    @property
    def text(self) -> google.protobuf.wrappers_pb2.StringValue:
        """Text represents a string inputs."""
    @property
    def json(self) -> google.protobuf.struct_pb2.Value:
        """JSON is represented by using google.protobuf.Value.
        see https://github.com/protocolbuffers/protobuf/blob/main/src/google/protobuf/struct.proto
        """
    @property
    def multipart(self) -> global___Multipart:
        """Multipart represents a multipart message.
        It comprises of a mapping from given type name to a subset of aforementioned types.
        """
    serialized_bytes: builtins.bytes
    """serialized_bytes is for data serialized in BentoML's internal serialization format."""
    def __init__(
        self,
        *,
        api_name: builtins.str = ...,
        ndarray: global___NDArray | None = ...,
        dataframe: global___DataFrame | None = ...,
        series: global___Series | None = ...,
        file: global___File | None = ...,
        text: google.protobuf.wrappers_pb2.StringValue | None = ...,
        json: google.protobuf.struct_pb2.Value | None = ...,
        multipart: global___Multipart | None = ...,
        serialized_bytes: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["content", b"content", "dataframe", b"dataframe", "file", b"file", "json", b"json", "multipart", b"multipart", "ndarray", b"ndarray", "serialized_bytes", b"serialized_bytes", "series", b"series", "text", b"text"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["api_name", b"api_name", "content", b"content", "dataframe", b"dataframe", "file", b"file", "json", b"json", "multipart", b"multipart", "ndarray", b"ndarray", "serialized_bytes", b"serialized_bytes", "series", b"series", "text", b"text"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["content", b"content"]) -> typing_extensions.Literal["ndarray", "dataframe", "series", "file", "text", "json", "multipart", "serialized_bytes"] | None: ...

global___Request = Request

@typing_extensions.final
class Response(google.protobuf.message.Message):
    """Request message for incoming Call."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NDARRAY_FIELD_NUMBER: builtins.int
    DATAFRAME_FIELD_NUMBER: builtins.int
    SERIES_FIELD_NUMBER: builtins.int
    FILE_FIELD_NUMBER: builtins.int
    TEXT_FIELD_NUMBER: builtins.int
    JSON_FIELD_NUMBER: builtins.int
    MULTIPART_FIELD_NUMBER: builtins.int
    SERIALIZED_BYTES_FIELD_NUMBER: builtins.int
    @property
    def ndarray(self) -> global___NDArray:
        """NDArray represents a n-dimensional array of arbitrary type."""
    @property
    def dataframe(self) -> global___DataFrame:
        """DataFrame represents any tabular data type. We are using
        DataFrame as a trivial representation for tabular type.
        """
    @property
    def series(self) -> global___Series:
        """Series portrays a series of values. This can be used for
        representing Series types in tabular data.
        """
    @property
    def file(self) -> global___File:
        """File represents for any arbitrary file type. This can be
        plaintext, image, video, audio, etc.
        """
    @property
    def text(self) -> google.protobuf.wrappers_pb2.StringValue:
        """Text represents a string inputs."""
    @property
    def json(self) -> google.protobuf.struct_pb2.Value:
        """JSON is represented by using google.protobuf.Value.
        see https://github.com/protocolbuffers/protobuf/blob/main/src/google/protobuf/struct.proto
        """
    @property
    def multipart(self) -> global___Multipart:
        """Multipart represents a multipart message.
        It comprises of a mapping from given type name to a subset of aforementioned types.
        """
    serialized_bytes: builtins.bytes
    """serialized_bytes is for data serialized in BentoML's internal serialization format."""
    def __init__(
        self,
        *,
        ndarray: global___NDArray | None = ...,
        dataframe: global___DataFrame | None = ...,
        series: global___Series | None = ...,
        file: global___File | None = ...,
        text: google.protobuf.wrappers_pb2.StringValue | None = ...,
        json: google.protobuf.struct_pb2.Value | None = ...,
        multipart: global___Multipart | None = ...,
        serialized_bytes: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["content", b"content", "dataframe", b"dataframe", "file", b"file", "json", b"json", "multipart", b"multipart", "ndarray", b"ndarray", "serialized_bytes", b"serialized_bytes", "series", b"series", "text", b"text"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["content", b"content", "dataframe", b"dataframe", "file", b"file", "json", b"json", "multipart", b"multipart", "ndarray", b"ndarray", "serialized_bytes", b"serialized_bytes", "series", b"series", "text", b"text"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["content", b"content"]) -> typing_extensions.Literal["ndarray", "dataframe", "series", "file", "text", "json", "multipart", "serialized_bytes"] | None: ...

global___Response = Response

@typing_extensions.final
class Part(google.protobuf.message.Message):
    """Part represents possible value types for multipart message.
    These are the same as the types in Request message.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NDARRAY_FIELD_NUMBER: builtins.int
    DATAFRAME_FIELD_NUMBER: builtins.int
    SERIES_FIELD_NUMBER: builtins.int
    FILE_FIELD_NUMBER: builtins.int
    TEXT_FIELD_NUMBER: builtins.int
    JSON_FIELD_NUMBER: builtins.int
    SERIALIZED_BYTES_FIELD_NUMBER: builtins.int
    @property
    def ndarray(self) -> global___NDArray:
        """NDArray represents a n-dimensional array of arbitrary type."""
    @property
    def dataframe(self) -> global___DataFrame:
        """DataFrame represents any tabular data type. We are using
        DataFrame as a trivial representation for tabular type.
        """
    @property
    def series(self) -> global___Series:
        """Series portrays a series of values. This can be used for
        representing Series types in tabular data.
        """
    @property
    def file(self) -> global___File:
        """File represents for any arbitrary file type. This can be
        plaintext, image, video, audio, etc.
        """
    @property
    def text(self) -> google.protobuf.wrappers_pb2.StringValue:
        """Text represents a string inputs."""
    @property
    def json(self) -> google.protobuf.struct_pb2.Value:
        """JSON is represented by using google.protobuf.Value.
        see https://github.com/protocolbuffers/protobuf/blob/main/src/google/protobuf/struct.proto
        """
    serialized_bytes: builtins.bytes
    """serialized_bytes is for data serialized in BentoML's internal serialization format."""
    def __init__(
        self,
        *,
        ndarray: global___NDArray | None = ...,
        dataframe: global___DataFrame | None = ...,
        series: global___Series | None = ...,
        file: global___File | None = ...,
        text: google.protobuf.wrappers_pb2.StringValue | None = ...,
        json: google.protobuf.struct_pb2.Value | None = ...,
        serialized_bytes: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["dataframe", b"dataframe", "file", b"file", "json", b"json", "ndarray", b"ndarray", "representation", b"representation", "serialized_bytes", b"serialized_bytes", "series", b"series", "text", b"text"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["dataframe", b"dataframe", "file", b"file", "json", b"json", "ndarray", b"ndarray", "representation", b"representation", "serialized_bytes", b"serialized_bytes", "series", b"series", "text", b"text"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["representation", b"representation"]) -> typing_extensions.Literal["ndarray", "dataframe", "series", "file", "text", "json", "serialized_bytes"] | None: ...

global___Part = Part

@typing_extensions.final
class Multipart(google.protobuf.message.Message):
    """Multipart represents a multipart message.
    It comprises of a mapping from given type name to a subset of aforementioned types.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class FieldsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___Part: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___Part | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    FIELDS_FIELD_NUMBER: builtins.int
    @property
    def fields(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Part]: ...
    def __init__(
        self,
        *,
        fields: collections.abc.Mapping[builtins.str, global___Part] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["fields", b"fields"]) -> None: ...

global___Multipart = Multipart

@typing_extensions.final
class File(google.protobuf.message.Message):
    """File represents for any arbitrary file type. This can be
    plaintext, image, video, audio, etc.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _FileType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _FileTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[File._FileType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        FILE_TYPE_UNSPECIFIED: File._FileType.ValueType  # 0
        FILE_TYPE_CSV: File._FileType.ValueType  # 1
        """file types"""
        FILE_TYPE_PLAINTEXT: File._FileType.ValueType  # 2
        FILE_TYPE_JSON: File._FileType.ValueType  # 3
        FILE_TYPE_BYTES: File._FileType.ValueType  # 4
        FILE_TYPE_PDF: File._FileType.ValueType  # 5
        FILE_TYPE_PNG: File._FileType.ValueType  # 6
        """image types"""
        FILE_TYPE_JPEG: File._FileType.ValueType  # 7
        FILE_TYPE_GIF: File._FileType.ValueType  # 8
        FILE_TYPE_BMP: File._FileType.ValueType  # 9
        FILE_TYPE_TIFF: File._FileType.ValueType  # 10
        FILE_TYPE_WEBP: File._FileType.ValueType  # 11
        FILE_TYPE_SVG: File._FileType.ValueType  # 12

    class FileType(_FileType, metaclass=_FileTypeEnumTypeWrapper):
        """FileType represents possible file type to be handled by BentoML.
        Currently, we only support plaintext (Text()), image (Image()), and file (File()).
        TODO: support audio and video streaming file types.
        """

    FILE_TYPE_UNSPECIFIED: File.FileType.ValueType  # 0
    FILE_TYPE_CSV: File.FileType.ValueType  # 1
    """file types"""
    FILE_TYPE_PLAINTEXT: File.FileType.ValueType  # 2
    FILE_TYPE_JSON: File.FileType.ValueType  # 3
    FILE_TYPE_BYTES: File.FileType.ValueType  # 4
    FILE_TYPE_PDF: File.FileType.ValueType  # 5
    FILE_TYPE_PNG: File.FileType.ValueType  # 6
    """image types"""
    FILE_TYPE_JPEG: File.FileType.ValueType  # 7
    FILE_TYPE_GIF: File.FileType.ValueType  # 8
    FILE_TYPE_BMP: File.FileType.ValueType  # 9
    FILE_TYPE_TIFF: File.FileType.ValueType  # 10
    FILE_TYPE_WEBP: File.FileType.ValueType  # 11
    FILE_TYPE_SVG: File.FileType.ValueType  # 12

    KIND_FIELD_NUMBER: builtins.int
    CONTENT_FIELD_NUMBER: builtins.int
    kind: global___File.FileType.ValueType
    """optional type of file, let it be csv, text, parquet, etc."""
    content: builtins.bytes
    """contents of file as bytes."""
    def __init__(
        self,
        *,
        kind: global___File.FileType.ValueType | None = ...,
        content: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_kind", b"_kind", "kind", b"kind"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_kind", b"_kind", "content", b"content", "kind", b"kind"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_kind", b"_kind"]) -> typing_extensions.Literal["kind"] | None: ...

global___File = File

@typing_extensions.final
class DataFrame(google.protobuf.message.Message):
    """DataFrame represents any tabular data type. We are using
    DataFrame as a trivial representation for tabular type.
    This message carries given implementation of tabular data based on given orientation.
    TODO: support index, records, etc.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COLUMN_NAMES_FIELD_NUMBER: builtins.int
    COLUMNS_FIELD_NUMBER: builtins.int
    @property
    def column_names(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """columns name"""
    @property
    def columns(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Series]:
        """columns orient.
        { column ↠ { index ↠ value } }
        """
    def __init__(
        self,
        *,
        column_names: collections.abc.Iterable[builtins.str] | None = ...,
        columns: collections.abc.Iterable[global___Series] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["column_names", b"column_names", "columns", b"columns"]) -> None: ...

global___DataFrame = DataFrame

@typing_extensions.final
class Series(google.protobuf.message.Message):
    """Series portrays a series of values. This can be used for
    representing Series types in tabular data.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BOOL_VALUES_FIELD_NUMBER: builtins.int
    FLOAT_VALUES_FIELD_NUMBER: builtins.int
    INT32_VALUES_FIELD_NUMBER: builtins.int
    INT64_VALUES_FIELD_NUMBER: builtins.int
    STRING_VALUES_FIELD_NUMBER: builtins.int
    DOUBLE_VALUES_FIELD_NUMBER: builtins.int
    @property
    def bool_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bool]:
        """A bool parameter value"""
    @property
    def float_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """A float parameter value"""
    @property
    def int32_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """A int32 parameter value"""
    @property
    def int64_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """A int64 parameter value"""
    @property
    def string_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """A string parameter value"""
    @property
    def double_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """represents a double parameter value."""
    def __init__(
        self,
        *,
        bool_values: collections.abc.Iterable[builtins.bool] | None = ...,
        float_values: collections.abc.Iterable[builtins.float] | None = ...,
        int32_values: collections.abc.Iterable[builtins.int] | None = ...,
        int64_values: collections.abc.Iterable[builtins.int] | None = ...,
        string_values: collections.abc.Iterable[builtins.str] | None = ...,
        double_values: collections.abc.Iterable[builtins.float] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["bool_values", b"bool_values", "double_values", b"double_values", "float_values", b"float_values", "int32_values", b"int32_values", "int64_values", b"int64_values", "string_values", b"string_values"]) -> None: ...

global___Series = Series

@typing_extensions.final
class NDArray(google.protobuf.message.Message):
    """NDArray represents a n-dimensional array of arbitrary type."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _DType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _DTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[NDArray._DType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        DTYPE_UNSPECIFIED: NDArray._DType.ValueType  # 0
        """Represents a None type."""
        DTYPE_FLOAT: NDArray._DType.ValueType  # 1
        """Represents an float type."""
        DTYPE_DOUBLE: NDArray._DType.ValueType  # 2
        """Represents an double type."""
        DTYPE_BOOL: NDArray._DType.ValueType  # 3
        """Represents a bool type."""
        DTYPE_INT32: NDArray._DType.ValueType  # 4
        """Represents an int32 type."""
        DTYPE_INT64: NDArray._DType.ValueType  # 5
        """Represents an int64 type."""
        DTYPE_UINT32: NDArray._DType.ValueType  # 6
        """Represents a uint32 type."""
        DTYPE_UINT64: NDArray._DType.ValueType  # 7
        """Represents a uint64 type."""
        DTYPE_STRING: NDArray._DType.ValueType  # 8
        """Represents a string type."""

    class DType(_DType, metaclass=_DTypeEnumTypeWrapper):
        """Represents data type of a given array."""

    DTYPE_UNSPECIFIED: NDArray.DType.ValueType  # 0
    """Represents a None type."""
    DTYPE_FLOAT: NDArray.DType.ValueType  # 1
    """Represents an float type."""
    DTYPE_DOUBLE: NDArray.DType.ValueType  # 2
    """Represents an double type."""
    DTYPE_BOOL: NDArray.DType.ValueType  # 3
    """Represents a bool type."""
    DTYPE_INT32: NDArray.DType.ValueType  # 4
    """Represents an int32 type."""
    DTYPE_INT64: NDArray.DType.ValueType  # 5
    """Represents an int64 type."""
    DTYPE_UINT32: NDArray.DType.ValueType  # 6
    """Represents a uint32 type."""
    DTYPE_UINT64: NDArray.DType.ValueType  # 7
    """Represents a uint64 type."""
    DTYPE_STRING: NDArray.DType.ValueType  # 8
    """Represents a string type."""

    DTYPE_FIELD_NUMBER: builtins.int
    SHAPE_FIELD_NUMBER: builtins.int
    STRING_VALUES_FIELD_NUMBER: builtins.int
    FLOAT_VALUES_FIELD_NUMBER: builtins.int
    DOUBLE_VALUES_FIELD_NUMBER: builtins.int
    BOOL_VALUES_FIELD_NUMBER: builtins.int
    INT32_VALUES_FIELD_NUMBER: builtins.int
    INT64_VALUES_FIELD_NUMBER: builtins.int
    UINT32_VALUES_FIELD_NUMBER: builtins.int
    UINT64_VALUES_FIELD_NUMBER: builtins.int
    dtype: global___NDArray.DType.ValueType
    """DTYPE is the data type of given array"""
    @property
    def shape(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """shape is the shape of given array."""
    @property
    def string_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """represents a string parameter value."""
    @property
    def float_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """represents a float parameter value."""
    @property
    def double_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """represents a double parameter value."""
    @property
    def bool_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bool]:
        """represents a bool parameter value."""
    @property
    def int32_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """represents a int32 parameter value."""
    @property
    def int64_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """represents a int64 parameter value."""
    @property
    def uint32_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """represents a uint32 parameter value."""
    @property
    def uint64_values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """represents a uint64 parameter value."""
    def __init__(
        self,
        *,
        dtype: global___NDArray.DType.ValueType = ...,
        shape: collections.abc.Iterable[builtins.int] | None = ...,
        string_values: collections.abc.Iterable[builtins.str] | None = ...,
        float_values: collections.abc.Iterable[builtins.float] | None = ...,
        double_values: collections.abc.Iterable[builtins.float] | None = ...,
        bool_values: collections.abc.Iterable[builtins.bool] | None = ...,
        int32_values: collections.abc.Iterable[builtins.int] | None = ...,
        int64_values: collections.abc.Iterable[builtins.int] | None = ...,
        uint32_values: collections.abc.Iterable[builtins.int] | None = ...,
        uint64_values: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["bool_values", b"bool_values", "double_values", b"double_values", "dtype", b"dtype", "float_values", b"float_values", "int32_values", b"int32_values", "int64_values", b"int64_values", "shape", b"shape", "string_values", b"string_values", "uint32_values", b"uint32_values", "uint64_values", b"uint64_values"]) -> None: ...

global___NDArray = NDArray

class BentoService(google.protobuf.service.Service, metaclass=abc.ABCMeta):
    """a gRPC BentoServer."""

    DESCRIPTOR: google.protobuf.descriptor.ServiceDescriptor
    @abc.abstractmethod
    def Call(
        inst: BentoService,  # pyright: ignore[reportSelfClsParameterName]
        rpc_controller: google.protobuf.service.RpcController,
        request: global___Request,
        callback: collections.abc.Callable[[global___Response], None] | None,
    ) -> concurrent.futures.Future[global___Response]:
        """Call handles methodcaller of given API entrypoint."""

class BentoService_Stub(BentoService):
    """a gRPC BentoServer."""

    def __init__(self, rpc_channel: google.protobuf.service.RpcChannel) -> None: ...
    DESCRIPTOR: google.protobuf.descriptor.ServiceDescriptor
    def Call(
        inst: BentoService_Stub,  # pyright: ignore[reportSelfClsParameterName]
        rpc_controller: google.protobuf.service.RpcController,
        request: global___Request,
        callback: collections.abc.Callable[[global___Response], None] | None = ...,
    ) -> concurrent.futures.Future[global___Response]:
        """Call handles methodcaller of given API entrypoint."""



from __future__ import annotations

import json
import os
import sys
import typing as t

import click


@click.command()
@click.argument("bento_identifier", type=click.STRING, required=False, default=".")
@click.option(
    "--runner-name",
    type=click.STRING,
    required=True,
    envvar="RUNNER_NAME",
)
@click.option("--fd", type=click.INT, required=True)
@click.option("--working-dir", required=False, default=None, help="Working directory")
@click.option(
    "--no-access-log",
    required=False,
    type=click.BOOL,
    is_flag=True,
    default=False,
    help="Disable the runner server's access log",
)
@click.option(
    "--worker-id",
    required=True,
    type=click.INT,
    help="If set, start the server as a bare worker with the given worker ID. Otherwise start a standalone server with a supervisor process.",
)
@click.option(
    "--worker-env-map",
    required=False,
    type=click.STRING,
    default=None,
    help="The environment variables to pass to the worker process. The format is a JSON string, e.g. '{0: {\"CUDA_VISIBLE_DEVICES\": 0}}'.",
)
@click.option(
    "--timeout",
    type=click.INT,
    help="Specify the timeout (seconds) for runner",
)
@click.option(
    "--prometheus-dir",
    type=click.Path(exists=True),
    help="Required by prometheus to pass the metrics in multi-process mode",
)
def main(
    bento_identifier: str,
    runner_name: str,
    fd: int,
    working_dir: t.Optional[str],
    no_access_log: bool,
    worker_id: int,
    timeout: int | None,
    worker_env_map: str | None,
    prometheus_dir: str | None,
) -> None:
    """
    Start a runner server.

    Args:
        bento_identifier: the Bento identifier
        name: the name of the runner
        fd: the file descriptor of the runner server's socket
        working_dir: (Optional) the working directory
        worker_id: (Optional) if set, the runner will be started as a worker with the given ID. Important: begin from 1.
        worker_env_map: (Optional) the environment variables to pass to the worker process. The format is a JSON string, e.g. '{0: {\"CUDA_VISIBLE_DEVICES\": 0}}'.
    """

    # setting up the environment for the worker process
    assert (
        "bentoml" not in sys.modules
    ), "bentoml should not be imported before setting up the environment, otherwise some of the environment may not take effect"
    if worker_env_map is not None:
        env_map: dict[str, dict[str, t.Any]] = json.loads(worker_env_map)
        worker_key = str(worker_id - 1)  # the worker ID is 1-based
        assert (
            worker_key in env_map
        ), f"worker_id {repr(worker_key)} not found in worker_env_map: {worker_env_map}"
        os.environ.update(env_map[worker_key])

    import socket

    import psutil

    from bentoml import load
    from bentoml._internal.context import component_context

    # setup context
    component_context.component_type = "runner"
    component_context.component_name = runner_name
    component_context.component_index = worker_id

    from bentoml._internal.log import configure_server_logging

    configure_server_logging()

    import uvicorn

    from bentoml._internal.configuration.containers import BentoMLContainer

    if prometheus_dir is not None:
        BentoMLContainer.prometheus_multiproc_dir.set(prometheus_dir)

    if no_access_log:
        access_log_config = BentoMLContainer.runners_config.logging.access
        access_log_config.enabled.set(False)

    from bentoml._internal.server.runner_app import RunnerAppFactory

    service = load(bento_identifier, working_dir=working_dir, standalone_load=True)

    # setup context
    if service.tag is None:
        component_context.bento_name = service.name
        component_context.bento_version = "not available"
    else:
        component_context.bento_name = service.tag.name
        component_context.bento_version = service.tag.version or "not available"

    for runner in service.runners:
        if runner.name == runner_name:
            break
    else:
        raise ValueError(f"Runner {runner_name} not found")

    factory = RunnerAppFactory(runner, worker_index=worker_id)
    if timeout is not None:
        factory.timeout = timeout
    app = factory()

    uvicorn_options: dict[str, int | None | str] = {
        "log_config": None,
        "workers": 1,
        "timeout_keep_alive": 1800,
    }

    if psutil.WINDOWS:
        # 1. uvloop is not supported on Windows
        # 2. the default policy for Python > 3.8 on Windows is ProactorEventLoop, which doesn't
        #    support listen on a existing socket file descriptors
        # See https://docs.python.org/3.8/library/asyncio-platforms.html#windows
        uvicorn_options["loop"] = "asyncio"
        import asyncio

        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())  # type: ignore

    # when fd is provided, we will skip the uvicorn internal supervisor, thus there is only one process
    sock = socket.socket(fileno=fd)
    config = uvicorn.Config(app, **uvicorn_options)
    uvicorn.Server(config).run(sockets=[sock])


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



from __future__ import annotations

import json
import logging
import os
import sys
from urllib.parse import urlparse

import click

logger = logging.getLogger(__name__)


def add_start_command(cli: click.Group) -> None:
    from bentoml._internal.configuration.containers import BentoMLContainer
    from bentoml._internal.utils import add_experimental_docstring
    from bentoml.grpc.utils import LATEST_PROTOCOL_VERSION

    @cli.command(hidden=True)
    @click.argument("bento", type=click.STRING, default=".")
    @click.option(
        "--remote-runner",
        type=click.STRING,
        multiple=True,
        envvar="BENTOML_SERVE_REMOTE_RUNNER",
        help="list of runners map",
    )
    @click.option(
        "--runner-map",
        type=click.STRING,
        envvar="BENTOML_SERVE_RUNNER_MAP",
        help="[Deprecated] use --remote-runner instead. "
        "JSON string of runners map. For backword compatibility for yatai < 1.0.0",
    )
    @click.option(
        "--bind",
        type=click.STRING,
        help="[Deprecated] use --host and --port instead."
        "Bind address for the server. For backword compatibility for yatai < 1.0.0",
        required=False,
    )
    @click.option(
        "--port",
        type=click.INT,
        default=BentoMLContainer.http.port.get(),
        help="The port to listen on for the REST api server",
        envvar="BENTOML_PORT",
        show_default=True,
    )
    @click.option(
        "--host",
        type=click.STRING,
        default=BentoMLContainer.http.host.get(),
        help="The host to bind for the REST api server [defaults: 127.0.0.1(dev), 0.0.0.0(production)]",
        envvar="BENTOML_HOST",
    )
    @click.option(
        "--backlog",
        type=click.INT,
        default=BentoMLContainer.api_server_config.backlog.get(),
        help="The maximum number of pending connections.",
        show_default=True,
    )
    @click.option(
        "--api-workers",
        type=click.INT,
        default=BentoMLContainer.api_server_workers.get(),
        help="Specify the number of API server workers to start. Default to number of available CPU cores in production mode",
        envvar="BENTOML_API_WORKERS",
    )
    @click.option(
        "--timeout",
        type=click.INT,
        help="Specify the timeout (seconds) for API server",
        envvar="BENTOML_TIMEOUT",
    )
    @click.option(
        "--working-dir",
        type=click.Path(),
        help="When loading from source code, specify the directory to find the Service instance",
        default=None,
        show_default=True,
    )
    @click.option(
        "--ssl-certfile",
        type=str,
        default=BentoMLContainer.ssl.certfile.get(),
        help="SSL certificate file",
    )
    @click.option(
        "--ssl-keyfile",
        type=str,
        default=BentoMLContainer.ssl.keyfile.get(),
        help="SSL key file",
    )
    @click.option(
        "--ssl-keyfile-password",
        type=str,
        default=BentoMLContainer.ssl.keyfile_password.get(),
        help="SSL keyfile password",
    )
    @click.option(
        "--ssl-version",
        type=int,
        default=BentoMLContainer.ssl.version.get(),
        help="SSL version to use (see stdlib 'ssl' module)",
    )
    @click.option(
        "--ssl-cert-reqs",
        type=int,
        default=BentoMLContainer.ssl.cert_reqs.get(),
        help="Whether client certificate is required (see stdlib 'ssl' module)",
    )
    @click.option(
        "--ssl-ca-certs",
        type=str,
        default=BentoMLContainer.ssl.ca_certs.get(),
        help="CA certificates file",
    )
    @click.option(
        "--ssl-ciphers",
        type=str,
        default=BentoMLContainer.ssl.ciphers.get(),
        help="Ciphers to use (see stdlib 'ssl' module)",
    )
    @add_experimental_docstring
    def start_http_server(  # type: ignore (unused warning)
        bento: str,
        remote_runner: list[str] | None,
        runner_map: str | None,
        bind: str | None,
        port: int,
        host: str,
        backlog: int,
        working_dir: str | None,
        api_workers: int | None,
        timeout: int | None,
        ssl_certfile: str | None,
        ssl_keyfile: str | None,
        ssl_keyfile_password: str | None,
        ssl_version: int | None,
        ssl_cert_reqs: int | None,
        ssl_ca_certs: str | None,
        ssl_ciphers: str | None,
    ) -> None:
        """
        Start a HTTP API server standalone. This will be used inside Yatai.
        """
        if working_dir is None:
            if os.path.isdir(os.path.expanduser(bento)):
                working_dir = os.path.expanduser(bento)
            else:
                working_dir = "."
        if sys.path[0] != working_dir:
            sys.path.insert(0, working_dir)

        from bentoml.start import start_http_server

        if remote_runner:
            runner_map_dict = dict(
                [s.split("=", maxsplit=2) for s in remote_runner or []]
            )
        elif runner_map:
            runner_map_dict = json.loads(runner_map)
        else:
            runner_map_dict = {}

        if bind is not None:
            parsed = urlparse(bind)
            assert parsed.scheme == "tcp"
            host = parsed.hostname or host
            port = parsed.port or port

        click.echo(f"Using remote runners: {runner_map}")
        start_http_server(
            bento,
            runner_map=runner_map_dict,
            working_dir=working_dir,
            port=port,
            host=host,
            backlog=backlog,
            api_workers=api_workers,
            timeout=timeout,
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
            ssl_keyfile_password=ssl_keyfile_password,
            ssl_version=ssl_version,
            ssl_cert_reqs=ssl_cert_reqs,
            ssl_ca_certs=ssl_ca_certs,
            ssl_ciphers=ssl_ciphers,
        )

    @cli.command(hidden=True)
    @click.argument("bento", type=click.STRING, default=".")
    @click.option(
        "--runner-name",
        type=click.STRING,
        required=True,
        envvar="BENTOML_SERVE_RUNNER_NAME",
        help="specify the runner name to serve",
    )
    @click.option(
        "--bind",
        type=click.STRING,
        help="[Deprecated] use --host and --port instead."
        "Bind address for the server. For backword compatibility for yatai < 1.0.0",
        required=False,
    )
    @click.option(
        "--port",
        type=click.INT,
        default=BentoMLContainer.http.port.get(),
        help="The port to listen on for the REST api server",
        envvar="BENTOML_PORT",
        show_default=True,
    )
    @click.option(
        "--host",
        type=click.STRING,
        default=BentoMLContainer.http.host.get(),
        help="The host to bind for the REST api server [defaults: 127.0.0.1(dev), 0.0.0.0(production)]",
        envvar="BENTOML_HOST",
    )
    @click.option(
        "--backlog",
        type=click.INT,
        default=BentoMLContainer.api_server_config.backlog.get(),
        help="The maximum number of pending connections.",
        show_default=True,
    )
    @click.option(
        "--working-dir",
        type=click.Path(),
        help="When loading from source code, specify the directory to find the Service instance",
        default=None,
        show_default=True,
    )
    @click.option(
        "--timeout",
        type=click.INT,
        help="Specify the timeout (seconds) for runners",
        envvar="BENTOML_TIMEOUT",
    )
    @add_experimental_docstring
    def start_runner_server(  # type: ignore (unused warning)
        bento: str,
        runner_name: str,
        bind: str | None,
        port: int,
        host: str,
        backlog: int,
        working_dir: str,
        timeout: int | None,
    ) -> None:
        """
        Start Runner server standalone. This will be used inside Yatai.
        """
        if working_dir is None:
            if os.path.isdir(os.path.expanduser(bento)):
                working_dir = os.path.expanduser(bento)
            else:
                working_dir = "."
        if sys.path[0] != working_dir:
            sys.path.insert(0, working_dir)

        from bentoml.start import start_runner_server

        if bind is not None:
            parsed = urlparse(bind)
            assert parsed.scheme == "tcp"
            host = parsed.hostname or host
            port = parsed.port or port

        start_runner_server(
            bento,
            runner_name=runner_name,
            working_dir=working_dir,
            timeout=timeout,
            port=port,
            host=host,
            backlog=backlog,
        )

    @cli.command(hidden=True)
    @click.argument("bento", type=click.STRING, default=".")
    @click.option(
        "--remote-runner",
        type=click.STRING,
        multiple=True,
        envvar="BENTOML_SERVE_RUNNER_MAP",
        help="JSON string of runners map",
    )
    @click.option(
        "--port",
        type=click.INT,
        default=BentoMLContainer.grpc.port.get(),
        help="The port to listen on for the gRPC server",
        envvar="BENTOML_PORT",
        show_default=True,
    )
    @click.option(
        "--host",
        type=click.STRING,
        default=BentoMLContainer.grpc.host.get(),
        help="The host to bind for the gRPC server (defaults: 0.0.0.0)",
        envvar="BENTOML_HOST",
    )
    @click.option(
        "--backlog",
        type=click.INT,
        default=BentoMLContainer.api_server_config.backlog.get(),
        help="The maximum number of pending connections.",
        show_default=True,
    )
    @click.option(
        "--working-dir",
        type=click.Path(),
        help="When loading from source code, specify the directory to find the Service instance",
        default=None,
        show_default=True,
    )
    @click.option(
        "--api-workers",
        type=click.INT,
        default=BentoMLContainer.api_server_workers.get(),
        help="Specify the number of API server workers to start. Default to number of available CPU cores in production mode",
        envvar="BENTOML_API_WORKERS",
    )
    @click.option(
        "--enable-reflection",
        is_flag=True,
        default=BentoMLContainer.grpc.reflection.enabled.get(),
        type=click.BOOL,
        help="Enable reflection.",
    )
    @click.option(
        "--enable-channelz",
        is_flag=True,
        default=BentoMLContainer.grpc.channelz.enabled.get(),
        type=click.BOOL,
        help="Enable Channelz. See https://github.com/grpc/proposal/blob/master/A14-channelz.md.",
    )
    @click.option(
        "--max-concurrent-streams",
        default=BentoMLContainer.grpc.max_concurrent_streams.get(),
        type=click.INT,
        help="Maximum number of concurrent incoming streams to allow on a http2 connection.",
    )
    @click.option(
        "--ssl-certfile",
        type=str,
        default=BentoMLContainer.ssl.certfile.get(),
        help="SSL certificate file",
    )
    @click.option(
        "--ssl-keyfile",
        type=str,
        default=BentoMLContainer.ssl.keyfile.get(),
        help="SSL key file",
    )
    @click.option(
        "--ssl-ca-certs",
        type=str,
        default=BentoMLContainer.ssl.ca_certs.get(),
        help="CA certificates file",
    )
    @click.option(
        "-pv",
        "--protocol-version",
        type=click.Choice(["v1", "v1alpha1"]),
        help="Determine the version of generated gRPC stubs to use.",
        default=LATEST_PROTOCOL_VERSION,
        show_default=True,
    )
    @add_experimental_docstring
    def start_grpc_server(  # type: ignore (unused warning)
        bento: str,
        remote_runner: list[str] | None,
        port: int,
        host: str,
        backlog: int,
        api_workers: int | None,
        working_dir: str,
        ssl_certfile: str | None,
        ssl_keyfile: str | None,
        enable_reflection: bool,
        ssl_ca_certs: str | None,
        enable_channelz: bool,
        max_concurrent_streams: int | None,
        protocol_version: str,
    ) -> None:
        """
        Start a gRPC API server standalone. This will be used inside Yatai.
        """
        if working_dir is None:
            if os.path.isdir(os.path.expanduser(bento)):
                working_dir = os.path.expanduser(bento)
            else:
                working_dir = "."
        if sys.path[0] != working_dir:
            sys.path.insert(0, working_dir)

        from bentoml.start import start_grpc_server

        runner_map = dict([s.split("=", maxsplit=2) for s in remote_runner or []])
        click.echo(f"Using remote runners: {runner_map}")
        start_grpc_server(
            bento,
            runner_map=runner_map,
            working_dir=working_dir,
            port=port,
            host=host,
            backlog=backlog,
            api_workers=api_workers,
            reflection=enable_reflection,
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
            ssl_ca_certs=ssl_ca_certs,
            channelz=enable_channelz,
            max_concurrent_streams=max_concurrent_streams,
            protocol_version=protocol_version,
        )



from __future__ import annotations

import functools
import logging
import os
import re
import sys
import typing as t
from shutil import which

import click
import fs
from simple_di import Provide
from simple_di import inject

from bentoml._internal.bento.bento import BENTO_YAML_FILENAME
from bentoml._internal.bento.bento import DEFAULT_BENTO_BUILD_FILE
from bentoml._internal.bento.bento import Bento
from bentoml._internal.bento.bento import BentoStore
from bentoml._internal.configuration import get_debug_mode
from bentoml._internal.configuration.containers import BentoMLContainer
from bentoml._internal.env_manager import EnvManager
from bentoml._internal.env_manager.envs import Environment
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import NotFound as BentoNotFound

if t.TYPE_CHECKING:
    P = t.ParamSpec("P")
    F = t.Callable[P, t.Any]

logger = logging.getLogger(__name__)


def remove_env_arg(cmd_args: list[str]) -> list[str]:
    """
    Removes `--env <env_name>` or `--env=<env_name>` from sys.argv
    """
    indices_to_remove: list[int] = []
    for i, arg in enumerate(cmd_args):
        # regex matching click.option
        if re.match(r"^--env=?.*", arg):
            indices_to_remove.append(i)
            if "=" not in arg:
                indices_to_remove.append(i + 1)

    new_cmd_args: list[str] = []
    for i, arg in enumerate(cmd_args):
        if i not in indices_to_remove:
            new_cmd_args.append(arg)
    return new_cmd_args


@inject
def get_environment(
    bento_identifier: str,
    env: t.Literal["conda"],
    bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
) -> Environment:
    # env created will be ephemeral
    if os.path.isdir(os.path.expanduser(bento_identifier)):
        bento_path_fs = fs.open_fs(
            os.path.abspath(os.path.expanduser(bento_identifier))
        )
        if bento_path_fs.isfile(BENTO_YAML_FILENAME):
            # path to a build bento dir
            return EnvManager.from_bento(
                env_type=env, bento=Bento.from_fs(bento_path_fs)
            ).environment
        elif bento_path_fs.isfile(DEFAULT_BENTO_BUILD_FILE):
            # path to a bento project
            raise NotImplementedError(
                "Serving from development project is not yet supported."
            )
        else:
            raise BentoMLException(
                f"EnvManager failed to create an environment from path {bento_path_fs}. When loading from a path, it must be either a Bento or a project directory containing '{DEFAULT_BENTO_BUILD_FILE}'."
            )
    else:
        try:
            bento = bento_store.get(bento_identifier)
            return EnvManager.from_bento(
                env_type=env,
                bento=bento,
                is_ephemeral=False,
            ).environment
        except BentoNotFound:
            # service definition
            raise NotImplementedError(
                "Serving bento with 'import_string' in an environment is not supported now."
            )


def env_manager(func: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
    @click.option(
        "--env",
        type=click.Choice(["conda"]),
        default=None,
        help="Environment to run the command in",
        show_default=True,
    )
    @functools.wraps(func)
    def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
        env = kwargs.pop("env")
        if env is not None:
            from rich.status import Status

            bento_identifier = kwargs["bento"]
            spinner_status = Status(f"Preparing {env} environment")
            if not get_debug_mode():
                spinner_status.start()
                bento_env = get_environment(bento_identifier, env)
                spinner_status.stop()
            else:
                bento_env = get_environment(bento_identifier, env)

            # once env is created, spin up a subprocess to run current arg
            bentoml_exec_path = which("bentoml")
            if bentoml_exec_path is None:
                raise BentoMLException("bentoml command not found!")
            bento_env.run([bentoml_exec_path] + remove_env_arg(sys.argv[1:]))

        return func(*args, **kwargs)

    return wrapper



from ._internal.frameworks.mlflow import MLFLOW_MODEL_FOLDER
from ._internal.frameworks.mlflow import get
from ._internal.frameworks.mlflow import get_mlflow_model
from ._internal.frameworks.mlflow import get_runnable
from ._internal.frameworks.mlflow import import_model
from ._internal.frameworks.mlflow import load_model

__all__ = [
    "load_model",
    "import_model",
    "get",
    "get_runnable",
    "get_mlflow_model",
    "MLFLOW_MODEL_FOLDER",
]



from __future__ import annotations

import typing as t
from typing import TYPE_CHECKING

import aiohttp
import multidict

if TYPE_CHECKING:
    from aiohttp.typedefs import LooseHeaders
    from starlette.datastructures import FormData
    from starlette.datastructures import Headers
    from starlette.types import Receive
    from starlette.types import Scope
    from starlette.types import Send


async def parse_multipart_form(headers: "Headers", body: bytes) -> "FormData":
    """
    parse starlette forms from headers and body
    """

    from starlette.formparsers import MultiPartParser

    async def async_bytesio(bytes_: bytes) -> t.AsyncGenerator[bytes, None]:
        yield bytes_
        yield b""
        return

    parser = MultiPartParser(headers=headers, stream=async_bytesio(body))
    return await parser.parse()


def handle_assert_exception(assert_object: t.Any, obj: t.Any, msg: str):
    res = assert_object
    try:
        if callable(assert_object):
            res = assert_object(obj)
            assert res
        else:
            assert obj == assert_object
    except AssertionError:
        raise ValueError(f"Expected: {res}. {msg}") from None
    except Exception as e:  # pylint: disable=broad-except
        # if callable has some errors, then we raise it here
        raise ValueError(
            f"Exception while excuting '{assert_object.__name__}': {e}"
        ) from None


async def async_request(
    method: str,
    url: str,
    headers: None | tuple[tuple[str, str], ...] | LooseHeaders = None,
    data: t.Any = None,
    timeout: int | None = None,
    assert_status: int | t.Callable[[int], bool] | None = None,
    assert_data: bytes | t.Callable[[bytes], bool] | None = None,
    assert_headers: t.Callable[[t.Any], bool] | None = None,
) -> tuple[int, Headers, bytes]:
    from starlette.datastructures import Headers

    async with aiohttp.ClientSession() as sess:
        try:
            async with sess.request(
                method, url, data=data, headers=headers, timeout=timeout
            ) as resp:
                body = await resp.read()
        except Exception:
            raise RuntimeError("Unable to reach host.") from None
    if assert_status is not None:
        handle_assert_exception(
            assert_status,
            resp.status,
            f"Return status [{resp.status}] with body: {body!r}",
        )
    if assert_data is not None:
        if callable(assert_data):
            msg = f"'{assert_data.__name__}' returns {assert_data(body)}"
        else:
            msg = f"Expects data '{assert_data}'"
        handle_assert_exception(
            assert_data,
            body,
            f"{msg}\nReceived response: {body}.",
        )
    if assert_headers is not None:
        handle_assert_exception(
            assert_headers,
            resp.headers,
            f"Headers assertion failed: {resp.headers!r}",
        )
    return resp.status, Headers(resp.headers), body


def assert_distributed_header(headers: multidict.CIMultiDict[str]) -> None:
    assert (
        headers.get("Yatai-Bento-Deployment-Name") == "test-deployment"
        and headers.get("Yatai-Bento-Deployment-Namespace") == "yatai"
    )


async def http_proxy_app(scope: Scope, receive: Receive, send: Send):
    """
    A simple HTTP proxy app that simulate the behavior of Yatai.
    """
    if scope["type"] == "lifespan":
        return

    if scope["type"] == "http":
        async with aiohttp.ClientSession() as session:
            headers = multidict.CIMultiDict(
                tuple((k.decode(), v.decode()) for k, v in scope["headers"])
            )

            assert_distributed_header(headers)
            bodies: list[bytes] = []
            while True:
                request_message = await receive()
                assert request_message["type"] == "http.request"
                request_body = request_message.get("body")
                assert isinstance(request_body, bytes)
                bodies.append(request_body)
                if not request_message["more_body"]:
                    break

            async with session.request(
                method=scope["method"],
                url=scope["path"],
                headers=headers,
                data=b"".join(bodies),
            ) as response:
                await send(
                    {
                        "type": "http.response.start",
                        "status": response.status,
                        "headers": list(response.raw_headers),
                    }
                )
                response_body: bytes = await response.read()
                await send(
                    {
                        "type": "http.response.body",
                        "body": response_body,
                        "more_body": False,
                    }
                )
        return

    raise NotImplementedError(f"Scope {scope} is not understood.") from None



from __future__ import annotations

import typing as t
from abc import ABC
from abc import abstractmethod
from functools import update_wrapper

from ...exceptions import InvalidArgument

if t.TYPE_CHECKING:
    from types import UnionType

    import pyarrow
    import pyspark.sql.types
    from starlette.requests import Request
    from starlette.responses import Response

    from bentoml.grpc.types import ProtoField

    from ..context import ServiceContext as Context
    from ..service.openapi.specification import Reference
    from ..service.openapi.specification import Schema
    from ..types import LazyType

    InputType = (
        UnionType
        | t.Type[t.Any]
        | LazyType[t.Any]
        | dict[str, t.Type[t.Any] | UnionType | LazyType[t.Any]]
    )
    OpenAPIResponse = dict[str, str | dict[str, t.Any]]
    F = t.Callable[..., t.Any]


IO_DESCRIPTOR_REGISTRY: dict[str, type[IODescriptor[t.Any]]] = {}

IOType = t.TypeVar("IOType")


def from_spec(spec: dict[str, t.Any]) -> IODescriptor[t.Any]:
    if "id" not in spec:
        raise InvalidArgument(f"IO descriptor spec ({spec}) missing ID.")

    if spec["id"] is None:
        raise InvalidArgument("No IO descriptor spec found.")

    return IO_DESCRIPTOR_REGISTRY[spec["id"]].from_spec(spec)


class _OpenAPIMeta:
    @abstractmethod
    def openapi_schema(self) -> Schema | Reference:
        raise NotImplementedError

    @abstractmethod
    def openapi_components(self) -> dict[str, t.Any] | None:
        raise NotImplementedError

    @abstractmethod
    def openapi_example(self) -> t.Any | None:
        raise NotImplementedError

    @abstractmethod
    def openapi_request_body(self) -> dict[str, t.Any]:
        raise NotImplementedError

    @abstractmethod
    def openapi_responses(self) -> dict[str, t.Any]:
        raise NotImplementedError


class IODescriptor(ABC, _OpenAPIMeta, t.Generic[IOType]):
    """
    IODescriptor describes the input/output data format of an InferenceAPI defined
    in a :code:`bentoml.Service`. This is an abstract base class for extending new HTTP
    endpoint IO descriptor types in BentoServer.
    """

    if t.TYPE_CHECKING:
        # Populated by subclasses. Makes pyright happy.
        def __init__(self, **kwargs: t.Any) -> None:
            ...

    __slots__ = (
        "_mime_type",
        "proto_fields",
        "descriptor_id",
        "from_sample",
        "_sample",
    )

    HTTP_METHODS = ("POST",)

    _mime_type: str
    descriptor_id: str | None
    proto_fields: tuple[ProtoField]

    def __init_subclass__(
        cls,
        *,
        descriptor_id: str | None = None,
        proto_fields: tuple[ProtoField] | None = None,
    ):
        if descriptor_id is not None:
            if descriptor_id in IO_DESCRIPTOR_REGISTRY:
                raise ValueError(
                    f"Descriptor ID {descriptor_id} already registered to {IO_DESCRIPTOR_REGISTRY[descriptor_id]}."
                )
            IO_DESCRIPTOR_REGISTRY[descriptor_id] = cls
        cls.descriptor_id = descriptor_id
        cls.proto_fields = proto_fields or t.cast("tuple[ProtoField]", ())

        cls._sample: IOType | None = None

        def _() -> F:
            def impl(
                cl_: type[IODescriptor[t.Any]], sample: t.Any, **kwargs: t.Any
            ) -> IODescriptor[t.Any]:
                klass = cl_(**kwargs)
                klass.sample = klass._from_sample(sample)
                return klass

            impl.__doc__ = cls._from_sample.__doc__
            return update_wrapper(impl, cls._from_sample)

        cls.from_sample = classmethod(_())

    def __repr__(self) -> str:
        return self.__class__.__qualname__

    @property
    def sample(self) -> IOType | None:
        return self._sample

    @sample.setter
    def sample(self, value: IOType) -> None:
        self._sample = value

    if t.TYPE_CHECKING:

        @classmethod
        def from_sample(cls, sample: t.Any, **attrs: t.Any) -> t.Self:
            ...

    @abstractmethod
    def _from_sample(self, sample: t.Any) -> IOType:
        """
        Creates a new instance of the IO Descriptor from given sample.

        Args:
            sample: The sample to create the instance from.
            **kwargs: Additional keyword arguments to pass to the constructor.

        Returns:
            An instance of the IODescriptor.
        """
        raise NotImplementedError

    @property
    def mime_type(self) -> str:
        return self._mime_type

    def to_spec(self) -> dict[str, t.Any] | None:
        return None

    @classmethod
    def from_spec(cls, spec: dict[str, t.Any]) -> t.Self:
        raise NotImplementedError

    @abstractmethod
    def input_type(self) -> InputType:
        raise NotImplementedError

    @abstractmethod
    async def from_http_request(self, request: Request) -> IOType:
        raise NotImplementedError

    @abstractmethod
    async def to_http_response(
        self, obj: IOType, ctx: Context | None = None
    ) -> Response:
        raise NotImplementedError

    @abstractmethod
    async def from_proto(self, field: t.Any) -> IOType:
        raise NotImplementedError

    @abstractmethod
    async def to_proto(self, obj: IOType) -> t.Any:
        raise NotImplementedError

    def from_arrow(self, batch: pyarrow.RecordBatch) -> IOType:
        raise NotImplementedError(
            "This IO descriptor does not currently support batch inference."
        )

    def to_arrow(self, obj: IOType) -> pyarrow.RecordBatch:
        raise NotImplementedError(
            "This IO descriptor does not currently support batch inference."
        )

    def spark_schema(self) -> pyspark.sql.types.StructType:
        raise NotImplementedError(
            "This IO descriptor does not currently support batch inference."
        )



from __future__ import annotations

import difflib
import functools
import logging
import os
import re
import time
import typing as t

import attr
import click
import click_option_group as cog
from click import ClickException
from click.exceptions import UsageError

if t.TYPE_CHECKING:
    from click import Command
    from click import Context
    from click import Group
    from click import HelpFormatter
    from click import Option
    from click import Parameter

    P = t.ParamSpec("P")

    F = t.Callable[P, t.Any]

    class ClickFunctionWrapper(t.Protocol[P]):
        __name__: str
        __click_params__: list[Option]

        def __call__(  # pylint: disable=no-method-argument
            *args: P.args,
            **kwargs: P.kwargs,
        ) -> F[P]:
            ...

    WrappedCLI = t.Callable[P, ClickFunctionWrapper[P]]

    ClickParamType = t.Sequence[t.Any] | bool | None

logger = logging.getLogger("bentoml")


def kwargs_transformers(
    f: F[t.Any] | None = None,
    *,
    transformer: F[t.Any],
    pass_click_context: bool = False,
) -> F[t.Any]:
    def decorator(_f: F[t.Any]) -> t.Callable[P, t.Any]:
        @functools.wraps(_f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> t.Any:
            transformed = {k: transformer(v) for k, v in kwargs.items()}
            if pass_click_context:
                return _f(click.get_current_context(), *args, **transformed)
            return _f(*args, **transformed)

        return wrapper

    if f is None:
        return decorator
    return decorator(f)


def _validate_docker_tag(tag: str) -> str:
    from bentoml.exceptions import BentoMLException

    if ":" in tag:
        name, version = tag.split(":")[:2]
    else:
        name, version = tag, None

    valid_name_pattern = re.compile(
        r"""
        ^(
        [a-z0-9]+      # alphanumeric
        (.|_{1,2}|-+)? # seperators
        )*$
        """,
        re.VERBOSE,
    )
    valid_version_pattern = re.compile(
        r"""
        ^
        [a-zA-Z0-9] # cant start with .-
        [ -~]{,127} # ascii match rest, cap at 128
        $
        """,
        re.VERBOSE,
    )

    if not valid_name_pattern.match(name):
        raise BentoMLException(
            f"Provided Docker Image tag {tag} is invalid. Name components may contain lowercase letters, digits and separators. A separator is defined as a period, one or two underscores, or one or more dashes."
        )
    if version and not valid_version_pattern.match(version):
        raise BentoMLException(
            f"Provided Docker Image tag {tag} is invalid. A tag name must be valid ASCII and may contain lowercase and uppercase letters, digits, underscores, periods and dashes. A tag name may not start with a period or a dash and may contain a maximum of 128 characters."
        )
    return tag


def validate_container_tag(
    ctx: Context, param: Parameter, tag: str | tuple[str] | None
) -> str | tuple[str] | None:
    from bentoml.exceptions import BentoMLException

    if not tag:
        return tag
    elif isinstance(tag, tuple):
        return tuple(map(_validate_docker_tag, tag))
    elif isinstance(tag, str):
        return _validate_docker_tag(tag)
    else:
        raise BentoMLException(f"Invalid tag type. Got {type(tag)}")


# NOTE: shim for bentoctl
def validate_docker_tag(
    ctx: Context, param: Parameter, tag: str | tuple[str] | None
) -> str | tuple[str] | None:
    logger.warning(
        "'validate_docker_tag' is now deprecated, use 'validate_container_tag' instead."
    )
    return validate_container_tag(ctx, param, tag)


@t.overload
def normalize_none_type(value: t.Mapping[str, t.Any]) -> t.Mapping[str, t.Any]:
    ...


@t.overload
def normalize_none_type(value: ClickParamType) -> ClickParamType:
    ...


def normalize_none_type(
    value: ClickParamType | t.Mapping[str, t.Any]
) -> ClickParamType | t.Mapping[str, t.Any]:
    if isinstance(value, (tuple, list, set, str)) and len(value) == 0:
        return
    if isinstance(value, dict):
        return {k: normalize_none_type(v) for k, v in value.items()}
    return value


def flatten_opt_tuple(value: t.Any) -> t.Any:
    from bentoml._internal.types import LazyType

    if LazyType["tuple[t.Any, ...]"](tuple).isinstance(value) and len(value) == 1:
        return value[0]
    return value


# NOTE: This is the key we use to store the transformed options in the CLI context.
MEMO_KEY = "_memoized"


def opt_callback(ctx: Context, param: Parameter, value: ClickParamType):
    # NOTE: our new options implementation will have the following format:
    #   --opt ARG[=|:]VALUE[,VALUE] (e.g., --opt key1=value1,value2 --opt key2:value3/value4:hello)
    # Argument and values per --opt has one-to-one or one-to-many relationship,
    # separated by '=' or ':'.
    # TODO: We might also want to support the following format:
    #  --opt arg1 value1 arg2 value2 --opt arg3 value3
    #  --opt arg1=value1,arg2=value2 --opt arg3:value3

    assert param.multiple, "Only use this callback when multiple=True."
    if MEMO_KEY not in ctx.params:
        # By default, multiple options are stored as a tuple.
        # our memoized options are stored as a dict.
        ctx.params[MEMO_KEY] = {}

    if param.name not in ctx.params[MEMO_KEY]:
        ctx.params[MEMO_KEY][param.name] = ()

    value = normalize_none_type(value)
    if value is not None and isinstance(value, tuple):
        for opt in value:
            o, *val = re.split(r"=|:", opt, maxsplit=1)
            norm = o.replace("-", "_")
            if len(val) == 0:
                # --opt bool
                ctx.params[MEMO_KEY][norm] = True
            else:
                # --opt key=value
                ctx.params[MEMO_KEY].setdefault(norm, ())
                ctx.params[MEMO_KEY][norm] += (*val,)
    return value


@attr.define
class SharedOptions:
    """This is the click.Context object that will be used in BentoML CLI."""

    cloud_context: str | None = attr.field(default=None)

    def with_options(self, **attrs: t.Any) -> t.Any:
        return attr.evolve(self, **attrs)


class BentoMLCommandGroup(click.Group):
    """
    Click command class customized for BentoML CLI, allow specifying a default
    command for each group defined.

    This command groups will also introduce support for aliases for commands and groups.

    Example:

    .. code-block:: python

        @click.group(cls=BentoMLCommandGroup)
        def cli(): ...

        @click.group(name="cloud", aliases=["cloud"], cls=BentoMLCommandGroup)
        def cli(): ...

        @cli.command(aliases=["serve-http"])
        def serve(): ...
    """

    NUMBER_OF_COMMON_PARAMS = 5  # NOTE: 4 shared options and a option group title

    @staticmethod
    def bentoml_common_params(func: F[P]) -> WrappedCLI[bool, bool, str | None]:
        # NOTE: update NUMBER_OF_COMMON_PARAMS when adding option.
        from bentoml._internal.configuration import DEBUG_ENV_VAR
        from bentoml._internal.configuration import QUIET_ENV_VAR
        from bentoml._internal.configuration import set_debug_mode
        from bentoml._internal.configuration import set_quiet_mode
        from bentoml._internal.log import configure_logging
        from bentoml._internal.utils.analytics import BENTOML_DO_NOT_TRACK

        @cog.optgroup.group("Global options")
        @cog.optgroup.option(
            "-q",
            "--quiet",
            is_flag=True,
            default=False,
            envvar=QUIET_ENV_VAR,
            help="Suppress all warnings and info logs",
        )
        @cog.optgroup.option(
            "--verbose",
            "--debug",
            "verbose",
            is_flag=True,
            default=False,
            envvar=DEBUG_ENV_VAR,
            help="Generate debug information",
        )
        @cog.optgroup.option(
            "--do-not-track",
            is_flag=True,
            default=False,
            envvar=BENTOML_DO_NOT_TRACK,
            help="Do not send usage info",
        )
        @cog.optgroup.option(
            "--context",
            "cloud_context",
            type=click.STRING,
            default=None,
            help="BentoCloud context name.",
        )
        @click.pass_context
        @functools.wraps(func)
        def wrapper(
            ctx: click.Context,
            quiet: bool,
            verbose: bool,
            cloud_context: str | None,
            *args: P.args,
            **kwargs: P.kwargs,
        ) -> t.Any:
            ctx.obj = SharedOptions(cloud_context=cloud_context)
            if quiet:
                set_quiet_mode(True)
                if verbose:
                    logger.warning("'--quiet' passed; ignoring '--verbose/--debug'")
            elif verbose:
                set_debug_mode(True)

            configure_logging()

            return func(*args, **kwargs)

        return wrapper

    @staticmethod
    def bentoml_track_usage(
        func: F[P] | WrappedCLI[bool, bool, str | None],
        cmd_group: click.Group,
        **kwargs: t.Any,
    ) -> WrappedCLI[bool]:
        from bentoml._internal.utils.analytics import BENTOML_DO_NOT_TRACK
        from bentoml._internal.utils.analytics import CliEvent
        from bentoml._internal.utils.analytics import cli_events_map
        from bentoml._internal.utils.analytics import track

        command_name = kwargs.get("name", func.__name__)

        @functools.wraps(func)
        def wrapper(do_not_track: bool, *args: P.args, **kwargs: P.kwargs) -> t.Any:
            if do_not_track:
                os.environ[BENTOML_DO_NOT_TRACK] = str(True)
                return func(*args, **kwargs)

            start_time = time.time_ns()

            def get_tracking_event(return_value: t.Any) -> CliEvent:
                # we only need to pass type checking here for group name
                # since we know the cmd_group.name to be BentoMLCommandGroup
                assert cmd_group.name

                if (
                    cmd_group.name in cli_events_map
                    and command_name in cli_events_map[cmd_group.name]
                ):
                    return cli_events_map[cmd_group.name][command_name](
                        cmd_group.name, command_name, return_value
                    )
                return CliEvent(
                    cmd_group=cmd_group.name,
                    cmd_name=command_name,
                )

            try:
                return_value = func(*args, **kwargs)
                event = get_tracking_event(return_value)
                duration_in_ns = time.time_ns() - start_time
                event.duration_in_ms = duration_in_ns / 1e6
                track(event)
                return return_value
            except BaseException as e:
                event = get_tracking_event(None)
                duration_in_ns = time.time_ns() - start_time
                event.duration_in_ms = duration_in_ns / 1e6
                event.error_type = type(e).__name__
                event.return_code = 2 if isinstance(e, KeyboardInterrupt) else 1
                track(event)
                raise

        return wrapper

    @staticmethod
    def raise_click_exception(
        func: F[P] | WrappedCLI[bool], cmd_group: click.Group, **kwargs: t.Any
    ) -> ClickFunctionWrapper[t.Any]:
        from bentoml._internal.configuration import get_debug_mode
        from bentoml.exceptions import BentoMLException

        command_name = kwargs.get("name", func.__name__)

        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> t.Any:
            try:
                return func(*args, **kwargs)
            except BentoMLException as err:
                msg = f"[{cmd_group.name}] `{command_name}` failed: {str(err)}"
                if get_debug_mode():
                    ClickException(click.style(msg, fg="red")).show()
                    raise err from None
                else:
                    raise ClickException(click.style(msg, fg="red")) from err

        return t.cast("ClickFunctionWrapper[t.Any]", wrapper)

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super(BentoMLCommandGroup, self).__init__(*args, **kwargs)
        # these two dictionaries will store known aliases for commands and groups
        self._commands: dict[str, list[str]] = {}
        self._aliases: dict[str, str] = {}

    def command(self, *args: t.Any, **kwargs: t.Any) -> t.Callable[[F[P]], Command]:
        if "context_settings" not in kwargs:
            kwargs["context_settings"] = {}
        kwargs["context_settings"]["max_content_width"] = 120
        aliases = kwargs.pop("aliases", None)

        def wrapper(func: F[P]) -> Command:
            # add common parameters to command.
            options = BentoMLCommandGroup.bentoml_common_params(func)
            # Send tracking events before command finish.
            usage = BentoMLCommandGroup.bentoml_track_usage(options, self, **kwargs)
            # If BentoMLException raise ClickException instead before exit.
            wrapped = BentoMLCommandGroup.raise_click_exception(usage, self, **kwargs)

            # move common parameters to end of the parameters list
            wrapped.__click_params__ = (
                wrapped.__click_params__[-self.NUMBER_OF_COMMON_PARAMS :]
                + wrapped.__click_params__[: -self.NUMBER_OF_COMMON_PARAMS]
            )
            cmd = super(BentoMLCommandGroup, self).command(*args, **kwargs)(wrapped)
            # add aliases to a given commands if it is specified.
            if aliases is not None:
                assert cmd.name
                self._commands[cmd.name] = aliases
                self._aliases.update({alias: cmd.name for alias in aliases})
            return cmd

        return wrapper

    def group(self, *args: t.Any, **kwargs: t.Any) -> t.Callable[[F[P]], Group]:
        aliases = kwargs.pop("aliases", None)

        def decorator(f: F[P]):
            # create the main group
            grp = super(BentoMLCommandGroup, self).group(*args, **kwargs)(f)

            if aliases is not None:
                assert grp.name
                self._commands[grp.name] = aliases
                self._aliases.update({k: grp.name for k in aliases})

            return grp

        return decorator

    def resolve_alias(self, cmd_name: str):
        return self._aliases[cmd_name] if cmd_name in self._aliases else cmd_name

    def get_command(self, ctx: Context, cmd_name: str) -> Command | None:
        cmd_name = self.resolve_alias(cmd_name)
        return super(BentoMLCommandGroup, self).get_command(ctx, cmd_name)

    def format_commands(self, ctx: Context, formatter: HelpFormatter) -> None:
        rows: list[tuple[str, str]] = []
        sub_commands = self.list_commands(ctx)

        max_len = max(len(cmd) for cmd in sub_commands)
        limit = formatter.width - 6 - max_len

        for sub_command in sub_commands:
            cmd = self.get_command(ctx, sub_command)
            if cmd is None:
                continue
            # If the command is hidden, then we skip it.
            if hasattr(cmd, "hidden") and cmd.hidden:
                continue
            if sub_command in self._commands:
                aliases = ", ".join(sorted(self._commands[sub_command]))
                sub_command = "%s (%s)" % (sub_command, aliases)
            # this cmd_help is available since click>=7
            # BentoML requires click>=7.
            cmd_help = cmd.get_short_help_str(limit)
            rows.append((sub_command, cmd_help))
        if rows:
            with formatter.section("Commands"):
                formatter.write_dl(rows)

    def resolve_command(
        self, ctx: Context, args: list[str]
    ) -> tuple[str | None, Command | None, list[str]]:
        try:
            return super(BentoMLCommandGroup, self).resolve_command(ctx, args)
        except UsageError as e:
            error_msg = str(e)
            original_cmd_name = click.utils.make_str(args[0])
            matches = difflib.get_close_matches(
                original_cmd_name, self.list_commands(ctx), 3, 0.5
            )
            if matches:
                fmt_matches = "\n    ".join(matches)
                error_msg += "\n\n"
                error_msg += f"Did you mean?\n    {fmt_matches}"
            raise UsageError(error_msg, e.ctx)


def is_valid_bento_tag(value: str) -> bool:
    try:
        from bentoml._internal.tag import Tag

        Tag.from_str(value)
        return True
    except ValueError:
        return False


def is_valid_bento_name(value: str) -> bool:
    return re.match(r"^[A-Za-z_0-9]*$", value) is not None



from __future__ import annotations

import json
import typing as t

import click
import click_option_group as cog

if t.TYPE_CHECKING:
    from .utils import SharedOptions


def add_cloud_command(cli: click.Group) -> click.Group:
    from bentoml._internal.cloud.client import RestApiClient
    from bentoml._internal.cloud.config import CloudClientConfig
    from bentoml._internal.cloud.config import CloudClientContext
    from bentoml._internal.cloud.config import add_context
    from bentoml._internal.cloud.config import default_context_name
    from bentoml._internal.utils import bentoml_cattr
    from bentoml.exceptions import CLIException
    from bentoml_cli.utils import BentoMLCommandGroup

    @cli.group(name="cloud", aliases=["yatai"], cls=BentoMLCommandGroup)
    def cloud():
        """BentoCloud Subcommands Groups

        \b
        Note that 'bentoml yatai' is mainly for backward compatible reason. It is recommended
        to use 'bentoml cloud'.
        """

    @cloud.command()
    @cog.optgroup.group(
        "Login", help="Required login options", cls=cog.RequiredAllOptionGroup
    )
    @cog.optgroup.option(
        "--endpoint",
        type=click.STRING,
        help="BentoCloud or Yatai endpoint, i.e: https://cloud.bentoml.com",
    )
    @cog.optgroup.option(
        "--api-token",
        type=click.STRING,
        help="BentoCloud or Yatai user API token",
    )
    @click.pass_obj
    def login(shared_options: SharedOptions, endpoint: str, api_token: str) -> None:  # type: ignore (not accessed)
        """Login to BentoCloud or Yatai server."""
        cloud_rest_client = RestApiClient(endpoint, api_token)
        user = cloud_rest_client.get_current_user()

        if user is None:
            raise CLIException("current user is not found")

        org = cloud_rest_client.get_current_organization()

        if org is None:
            raise CLIException("current organization is not found")

        ctx = CloudClientContext(
            name=shared_options.cloud_context
            if shared_options.cloud_context is not None
            else default_context_name,
            endpoint=endpoint,
            api_token=api_token,
            email=user.email,
        )

        add_context(ctx)
        click.echo(
            f"Successfully logged in to BentoCloud for {user.name} in {org.name}"
        )

    @cloud.command()
    def current_context() -> None:  # type: ignore (not accessed)
        """Get current cloud context."""
        click.echo(
            json.dumps(
                bentoml_cattr.unstructure(
                    CloudClientConfig.get_config().get_current_context()
                ),
                indent=2,
            )
        )

    @cloud.command()
    def list_context() -> None:  # type: ignore (not accessed)
        """List all available context."""
        config = CloudClientConfig.get_config()
        click.echo(
            json.dumps(
                bentoml_cattr.unstructure([i.name for i in config.contexts]), indent=2
            )
        )

    @cloud.command()
    @click.argument("context_name", type=click.STRING)
    def update_current_context(context_name: str) -> None:  # type: ignore (not accessed)
        """Update current context"""
        ctx = CloudClientConfig.get_config().set_current_context(context_name)
        click.echo(f"Successfully switched to context: {ctx.name}")

    return cli



from __future__ import annotations

import logging
import typing as t
from typing import TYPE_CHECKING

from ....exceptions import BentoMLException
from ...types import LazyType
from ...utils.lazy_loader import LazyLoader

if TYPE_CHECKING:
    import onnx
    import torch

    from ... import external_typing as ext
    from ...external_typing import tensorflow as tf_ext

    ONNXArgTensorType = (
        ext.NpNDArray
        | ext.PdDataFrame
        | torch.Tensor
        | tf_ext.Tensor
        | list[int | float | str]
    )
    ONNXArgSequenceType = list["ONNXArgType"]
    ONNXArgMapKeyType = int | str
    ONNXArgMapType = dict[ONNXArgMapKeyType, "ONNXArgType"]
    ONNXArgType = ONNXArgMapType | ONNXArgTensorType | ONNXArgSequenceType

    ONNXArgCastedType = (
        ext.NpNDArray
        | list["ONNXArgCastedType"]
        | dict[ONNXArgMapKeyType, "ONNXArgCastedType"]
    )
    ONNXArgCastingFuncType = t.Callable[[ONNXArgType], ONNXArgCastedType]
    ONNXArgCastingFuncGeneratorType = t.Callable[
        [dict[str, t.Any]], t.Callable[[ONNXArgType], ONNXArgCastedType]
    ]

else:
    np = LazyLoader("np", globals(), "numpy")
    onnx = LazyLoader(
        "onnx",
        globals(),
        "onnx",
        exc_msg="`onnx` is required to use bentoml.onnx module.",
    )

logger = logging.getLogger(__name__)

TENSORPROTO_ELEMENT_TYPE_TO_NUMPY_TYPE: dict[int, str] = {
    onnx.TensorProto.FLOAT: "float32",  # 1
    onnx.TensorProto.UINT8: "uint8",  # 2
    onnx.TensorProto.INT8: "int8",  # 3
    onnx.TensorProto.UINT16: "uint16",  # 4
    onnx.TensorProto.INT16: "int16",  # 5
    onnx.TensorProto.INT32: "int32",  # 6
    onnx.TensorProto.INT64: "int64",  # 7
    onnx.TensorProto.STRING: "str",  # 8 or "unicode"?
    onnx.TensorProto.BOOL: "bool",  # 9
    onnx.TensorProto.FLOAT16: "float16",  # 10
    onnx.TensorProto.DOUBLE: "double",  # 11
    onnx.TensorProto.UINT32: "uint32",  # 12
    onnx.TensorProto.UINT64: "uint64",  # 13
    onnx.TensorProto.COMPLEX64: "csingle",  # 14
    onnx.TensorProto.COMPLEX128: "cdouble",  # 15
    # onnx.TensorProto.BFLOAT16: None,  # 16
}

# type -> casting function generator
CASTING_FUNC_DISPATCHER: dict[str, ONNXArgCastingFuncGeneratorType] = {}


def gen_input_casting_func(spec: dict[str, t.Any]) -> ONNXArgCastingFuncType:
    return _gen_input_casting_func(spec["type"])


def _gen_input_casting_func(sig: dict[str, t.Any]) -> ONNXArgCastingFuncType:
    input_types = list(sig.keys())
    if len(input_types) != 1:
        raise BentoMLException(
            "onnx model input type dictionary should have only one key!"
        )
    input_type = input_types[0]
    input_spec = sig[input_type]
    return CASTING_FUNC_DISPATCHER[input_type](input_spec)


def _gen_input_casting_func_for_tensor(
    sig: dict[str, t.Any]
) -> t.Callable[[ONNXArgTensorType], ext.NpNDArray]:
    elem_type = sig["elemType"]
    to_dtype = TENSORPROTO_ELEMENT_TYPE_TO_NUMPY_TYPE[elem_type]

    def _mapping(item: ONNXArgTensorType) -> ext.NpNDArray:
        if LazyType["ext.NpNDArray"]("numpy.ndarray").isinstance(item):
            item = item.astype(to_dtype, copy=False)
        elif isinstance(item, list):
            item = np.array(item).astype(to_dtype, copy=False)
        elif LazyType["ext.PdDataFrame"]("pandas.DataFrame").isinstance(item):
            item = item.to_numpy(dtype=to_dtype)
        elif LazyType["tf.Tensor"]("tensorflow.Tensor").isinstance(item):
            item = np.array(memoryview(item)).astype(to_dtype, copy=False)  # type: ignore
        elif LazyType["torch.Tensor"]("torch.Tensor").isinstance(item):
            item = item.numpy().astype(to_dtype, copy=False)
        else:
            raise TypeError(
                "`run` of ONNXRunnable only takes `numpy.ndarray`, `pd.DataFrame`, `tf.Tensor`, `torch.Tensor` or a list as input Tensor type"
            )
        return t.cast("ext.NpNDArray", item)

    return _mapping


CASTING_FUNC_DISPATCHER["tensorType"] = t.cast(
    "ONNXArgCastingFuncGeneratorType", _gen_input_casting_func_for_tensor
)


def _gen_input_casting_func_for_map(
    sig: dict[str, t.Any]
) -> t.Callable[[ONNXArgMapType], dict[ONNXArgMapKeyType, ONNXArgCastedType]]:
    map_value_sig = t.cast(dict[str, t.Any], sig["valueType"])
    value_casting_func = _gen_input_casting_func(map_value_sig)

    def _mapping(item: ONNXArgMapType) -> dict[ONNXArgMapKeyType, t.Any]:
        new_item = {k: value_casting_func(v) for k, v in item.items()}
        return new_item

    return _mapping


CASTING_FUNC_DISPATCHER["mapType"] = t.cast(
    "ONNXArgCastingFuncGeneratorType", _gen_input_casting_func_for_map
)


def _gen_input_casting_func_for_sequence(
    sig: dict[str, t.Any]
) -> t.Callable[[ONNXArgSequenceType], list[t.Any]]:
    seq_elem_sig = t.cast(dict[str, t.Any], sig["elemType"])
    elem_casting_func = _gen_input_casting_func(seq_elem_sig)

    def _mapping(item: ONNXArgSequenceType) -> list[t.Any]:
        new_item = list(elem_casting_func(elem) for elem in item)
        return new_item

    return _mapping


CASTING_FUNC_DISPATCHER["sequenceType"] = t.cast(
    "ONNXArgCastingFuncGeneratorType", _gen_input_casting_func_for_sequence
)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import bentoml.grpc.v1.service_pb2
import grpc

class BentoServiceStub:
    """a gRPC BentoServer."""

    def __init__(self, channel: grpc.Channel) -> None: ...
    Call: grpc.UnaryUnaryMultiCallable[
        bentoml.grpc.v1.service_pb2.Request,
        bentoml.grpc.v1.service_pb2.Response,
    ]
    """Call handles methodcaller of given API entrypoint."""
    ServiceMetadata: grpc.UnaryUnaryMultiCallable[
        bentoml.grpc.v1.service_pb2.ServiceMetadataRequest,
        bentoml.grpc.v1.service_pb2.ServiceMetadataResponse,
    ]
    """ServiceMetadata returns metadata of bentoml.Service."""

class BentoServiceServicer(metaclass=abc.ABCMeta):
    """a gRPC BentoServer."""

    @abc.abstractmethod
    def Call(
        self,
        request: bentoml.grpc.v1.service_pb2.Request,
        context: grpc.ServicerContext,
    ) -> bentoml.grpc.v1.service_pb2.Response:
        """Call handles methodcaller of given API entrypoint."""
    @abc.abstractmethod
    def ServiceMetadata(
        self,
        request: bentoml.grpc.v1.service_pb2.ServiceMetadataRequest,
        context: grpc.ServicerContext,
    ) -> bentoml.grpc.v1.service_pb2.ServiceMetadataResponse:
        """ServiceMetadata returns metadata of bentoml.Service."""

def add_BentoServiceServicer_to_server(servicer: BentoServiceServicer, server: grpc.Server) -> None: ...



"""
This module is shim for bentoctl. NOT FOR DIRECT USE.
Make sure to use 'bentoml.container.build' and 'bentoml.container.health' instead.
"""

from __future__ import annotations

import typing as t
import warnings

from ..container import get_backend
from ..container import health as _internal_container_health

__all__ = ["build", "health"]

_buildx_backend = get_backend("buildx")

warnings.warn(
    "%s is deprecated. Make sure to use 'bentoml.container.build' and 'bentoml.container.health' instead."
    % __name__,
    DeprecationWarning,
    stacklevel=4,
)


def health():
    return _internal_container_health("buildx")


def build(**kwargs: t.Any):
    # subprocess_env from bentoctl will be handle by buildx, so it is safe to pop this out.
    kwargs.pop("subprocess_env")
    kwargs["tag"] = kwargs.pop("tags")
    context_path = kwargs.pop("cwd", None)
    for key, value in kwargs.items():
        if not value:
            kwargs[key] = None
    _buildx_backend.build(context_path=context_path, **kwargs)
